from typing import Dict, Optional
from retrying import retry

from xtb_config.frozen import Frozen
from xtb_config.exception import OrderPendingException, \
    retry_if_order_not_sent_exception, retry_if_order_not_accepted_exception
from xtb_config.constants import Type, CMD, Cmd, STATUS, Status
from xtb_config.logger import logger
from xtb import Xtb
from xtb_models.transaction import Transaction


class Trade(Frozen):
    def __init__(self, trade_raw: Dict):
        trade_raw['order'] = trade_raw.get('order', 0)
        trade_raw['offset'] = trade_raw.get('offset', 0)
        trade_raw['sl'] = trade_raw.get('sl', 0)
        trade_raw['tp'] = trade_raw.get('tp', 0)
        trade_raw['profit'] = trade_raw.get('profit', 0)
        trade_raw['expiration'] = trade_raw.get('expiration', 0)
        trade_raw['storage'] = trade_raw.get('storage', 0)
        self.__order: Optional[int] = trade_raw['order']
        self.__cmd: int = trade_raw['cmd']
        self.__comment: str = trade_raw['comment']
        self.__offset: Optional[float] = trade_raw['offset']
        self.__sl: Optional[float] = trade_raw['sl']
        self.__tp: Optional[float] = trade_raw['tp']
        self.__symbol: str = trade_raw['symbol']
        self.__volume: float = trade_raw['volume']
        self.__profit: Optional[float] = trade_raw['profit']
        self.__storage: Optional[float] = trade_raw['storage']
        self.__raw: Dict = trade_raw
        self.__transaction: Optional[Transaction] = None


    @property
    def order(self) -> int:
        return self.__order

    # TODO: la mayoría de los setters sobran
    @order.setter
    def order(self, value: int) -> None:
        self.__order = value

    @property
    def cmd(self) -> int:
        return self.__cmd

    @cmd.setter
    def cmd(self, value: int) -> None:
        self.__cmd = value

    @property
    def comment(self) -> str:
        return self.__comment

    @comment.setter
    def comment(self, value: str) -> None:
        self.__comment = value

    @property
    def offset(self) -> float:
        return self.__offset

    @offset.setter
    def offset(self, value: float) -> None:
        self.__offset = value

    @property
    def sl(self) -> float:
        return self.__sl

    @sl.setter
    def sl(self, value: float) -> None:
        self.__sl = value

    @property
    def tp(self) -> float:
        return self.__tp

    @tp.setter
    def tp(self, value: float) -> None:
        self.__tp = value

    @property
    def symbol(self) -> str:
        return self.__symbol

    @symbol.setter
    def symbol(self, value: str) -> None:
        self.__symbol = value

    @property
    def volume(self) -> float:
        return self.__volume

    @volume.setter
    def volume(self, value: float) -> None:
        self.__volume = value

    @property
    def raw(self) -> Dict:
        return self.__raw

    @raw.setter
    def raw(self, value: Dict) -> None:
        self.__raw = value

    @property
    def profit(self) -> float:
        return self.__profit

    @profit.setter
    def profit(self, value: float) -> None:
        self.__profit = value

    @property
    def storage(self) -> float:
        return self.__storage

    @storage.setter
    def storage(self, value: float) -> None:
        self.__storage = value

    @property
    def transaction(self) -> Transaction:
        return self.__transaction

    @transaction.setter
    def transaction(self, value: Transaction) -> None:
        self.__transaction = value

    @property
    def accepted(self) -> bool:
        try:
            return self.transaction.accepted
        except:
            return False

    def transact(self, xtb: Xtb, _type: int) -> None:
        symbol = xtb.get_client().get_symbol(symbol=self.__symbol)
        if _type == Type.OPEN:
            if self.cmd == Cmd.BUY:
                self.raw['open_price'] = symbol['ask']
            elif self.cmd == Cmd.SELL:
                self.raw['open_price'] = symbol['bid']
        self.send_transaction(xtb=xtb, _type=_type)
        self.check_order_sent(xtb=xtb)

    @retry(retry_on_exception=retry_if_order_not_sent_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def send_transaction(self, xtb: Xtb, _type: int) -> None:
        # TODO: check lotMin del symbol antes de mandar transaction y devolver una exception
        self.transaction = Transaction(order=xtb.get_client().trade_transaction(trade=self.raw, _type=_type))

    @retry(retry_on_exception=retry_if_order_not_accepted_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def check_order_sent(self, xtb: Xtb):
        try:
            transaction = xtb.get_client().trade_transaction_status(transacting_order=self.transaction.order)
            self.transaction.status = STATUS[transaction['requestStatus']]
            self.transaction.message = transaction['message']
            self.transaction.custom_comment = transaction['customComment']
            self.transaction.ask = transaction['ask']
            self.transaction.bid = transaction['bid']
            if self.transaction.status == Status.ACCEPTED:
                logger.debug(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' is CLOSED. "
                             f"Transaction order is {self.transaction.order}")
            elif self.transaction.status == Status.PENDING:
                logger.warning(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' is PENDING. "
                               f"Transaction order is {self.transaction}")
                raise OrderPendingException(f"Trade order '{self.order} is PENDING. "
                                            f"Closing order is {self.transaction}")
            elif self.transaction.status == Status.ERROR:
                logger.error(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' has FAILED. "
                             f"Transaction: {self.transaction}")
            elif self.transaction.status == Status.REJECTED:
                logger.error(f"Trade order '{self.order} - {self.symbol} - {CMD[self.cmd]}' has been REJECTED. "
                             f"Transaction order is {self.transaction}")
        except OrderPendingException:
            self.transaction.status = Status.PENDING
        except Exception as e:
            logger.error(e)



from typing import Optional, Dict

from xtb_config.frozen import Frozen
from xtb_config.constants import Status


class Transaction(Frozen):
    def __init__(self, order: int):
        self.__order: int = order
        self.__status: str = ''
        self.__message: str = ''
        self.__custom_comment: Optional[str] = None
        self.__ask: float = 0.0
        self.__bid: float = 0.0

    def __repr__(self) -> str:
        return str(self.__to_dict())

    def __to_dict(self) -> Dict:
        return dict(
            order=self.__order,
            status=self.__status,
            message=self.__message,
            custom_comment=self.__custom_comment,
            ask=self.__ask,
            bid=self.__bid)

    @property
    def order(self) -> int:
        return self.__order

    @order.setter
    def order(self, value: int) -> None:
        self.__order = value

    @property
    def status(self) -> int:
        return self.__status

    @status.setter
    def status(self, value: int) -> None:
        self.__status = value

    @property
    def message(self) -> str:
        return self.__message

    @message.setter
    def message(self, value: str) -> None:
        self.__message = value

    @property
    def custom_comment(self) -> str:
        return self.__custom_comment

    @custom_comment.setter
    def custom_comment(self, value: str) -> None:
        self.__custom_comment = value

    @property
    def ask(self) -> float:
        return self.__ask

    @ask.setter
    def ask(self, value: float) -> None:
        self.__ask = value

    @property
    def bid(self) -> float:
        return self.__bid

    @bid.setter
    def bid(self, value: float) -> None:
        self.__bid = value

    @property
    def accepted(self) -> bool:
        return self.status == Status.ACCEPTED


from typing import Dict, Optional
from datetime import datetime

from xtb_config.frozen import Frozen
from xtb_config.constants import WEEKDAY
from xtb_models.timetable import Timetable
from xtb_models.shift import Shift

from xtb import Xtb



class Symbol(Frozen):
    def __init__(self, xtb: Xtb, symbol: str):
        symbol_dict: Dict = xtb.get_client().get_symbol(symbol=symbol)
        self.__symbol: str = symbol_dict['symbol']
        self.__currency: str = symbol_dict['currency']
        self.__currency_profit: str = symbol_dict['currencyProfit']
        self.__contract_size: str = symbol_dict['contractSize']
        self.__precision: str = symbol_dict['precision']
        self.__bid: str = symbol_dict['bid']
        self.__ask: str = symbol_dict['ask']
        self.__swap_long: str = symbol_dict['swapLong']
        self.__swap_short: str = symbol_dict['swapShort']
        self.__spread: float = symbol_dict['spreadRaw']
        self.__timetable: Timetable = xtb.get_symbol_timetable(symbol=symbol)
        
    @property
    def symbol(self) -> str:
        return self.__symbol

    @symbol.setter
    def symbol(self, value: str) -> None:
        self.__symbol = value

    @property
    def currency(self) -> str:
        return self.__currency

    @currency.setter
    def currency(self, value: str) -> None:
        self.__currency = value

    @property
    def currency_profit(self) -> str:
        return self.__currency_profit

    @property
    def contract_size(self) -> str:
        return self.__contract_size

    @contract_size.setter
    def contract_size(self, value: str) -> None:
        self.__contract_size = value

    @property
    def precision(self) -> str:
        return self.__precision

    @precision.setter
    def precision(self, value: str) -> None:
        self.__precision = value

    @property
    def bid(self) -> str:
        return self.__bid

    @bid.setter
    def bid(self, value: str) -> None:
        self.__bid = value

    @property
    def ask(self) -> str:
        return self.__ask

    @ask.setter
    def ask(self, value: str) -> None:
        self.__ask = value

    @property
    def swap_long(self) -> str:
        return self.__swap_long

    @swap_long.setter
    def swap_long(self, value: str) -> None:
        self.__swap_long = value

    @property
    def swap_short(self) -> str:
        return self.__swap_short

    @swap_short.setter
    def swap_short(self, value: str) -> None:
        self.__swap_short = value

    @property
    def timetable(self) -> Timetable:
        return self.__timetable

    @timetable.setter
    def timetable(self, value: Timetable) -> None:
        self.__timetable = value

    @property
    def spread(self) -> float:
        return self.__spread

    @spread.setter
    def spread(self, value: float) -> None:
        self.__spread = value

    def get_active_shift(self) -> Optional[Shift]:
        for shift in getattr(self.timetable, WEEKDAY[datetime.today().isoweekday()]):
            if shift.from_ts < datetime.now() < shift.to_ts:
                return shift



from typing import Dict, Optional

from xtb_config.frozen import Frozen
from xtb_config.constants import Cmd
from xtb_models.symbol import Symbol
from xtb_models.shift import Shift
from xtb_models.trade import Trade

from xtb import Xtb


class Position(Frozen):
    def __init__(self,
                 xtb: Xtb,
                 order: int,
                 direction: str,
                 symbol: str,
                 volume: float,
                 trade: Dict):
        self.__direction: int = getattr(Cmd, direction)
        self.__order: int = order
        self.__symbol: Symbol = Symbol(xtb=xtb, symbol=symbol)
        self.__trade: Trade = Trade(trade_raw=trade)
        self.__shift: Optional[Shift] = self.__symbol.get_active_shift()
        self.__volume: float = volume
        self.__action: str = ""

    @property
    def direction(self) -> int:
        return self.__direction

    @property
    def order(self) -> int:
        return self.__order

    @order.setter
    def order(self, value: int) -> None:
        self.__order = value

    @property
    def trade(self) -> Trade:
        return self.__trade

    @trade.setter
    def trade(self, value: Dict) -> None:
        self.__frozen__ = False
        self.__trade = value
        self.__frozen__ = True

    @property
    def symbol(self) -> Symbol:
        return self.__symbol

    @property
    def shift(self) -> Shift:
        return self.__shift

    @shift.setter
    def shift(self, value: Shift) -> None:
        self.__shift = value

    @property
    def volume(self) -> float:
        return self.__volume

    @volume.setter
    def volume(self, value: float) -> None:
        self.__volume = value

    @property
    def is_working(self) -> bool:
        return bool(self.__shift)

    @property
    def action(self) -> str:
        return self.__action

    @action.setter
    def action(self, value: str) -> None:
        self.__action = value

    @property
    def name(self) -> str:
        return f"{self.symbol} {self.direction} {self.volume}"

    @property
    def profit(self) -> float:
        if self.trade:
            return round(self.trade.profit + self.trade.storage, 2)
        return 0



class FrozenError(AttributeError):
    pass


class GenericException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__}, {self.message}"


class GetSymbolException(GenericException):
    pass


class GetMarginLevel(GenericException):
    pass


class GetMarginTrade(GenericException):
    pass


class LoginException(GenericException):
    pass


class GetTradeException(GenericException):
    pass


class GetTickPricesException(GenericException):
    pass


def retry_if_get_tick_prices_exception(exception):
    # logger.warning(f"{exception}")
    return isinstance(exception, GetTickPricesException)


class GetTradingHoursException(GenericException):
    pass


class GetTradeRecordsException(GenericException):
    pass


class OrderNotSentException(GenericException):
    pass


def retry_if_order_not_sent_exception(exception):
    # logger.warning(f"{exception}")
    return isinstance(exception, OrderNotSentException)


class OrderPendingException(GenericException):
    pass


def retry_if_order_not_accepted_exception(exception):
    # logger.warning(f"{exception}")
    return isinstance(exception, OrderPendingException) or isinstance(exception, OrderStatusNotSentException)


class OrderStatusNotSentException(GenericException):
    pass


class ArbitragePartiallyClosedException(GenericException):
    pass


def retry_if_arbitrage_partially_closed_exception(exception):
    # logger.warning(f"{exception}")
    return isinstance(exception, ArbitragePartiallyClosedException)


class TradeNotFoundException(GenericException):
    pass


class NotReceivedDataError(GenericException):
    pass



from xtb_config.exception import FrozenError


class Frozen(object):
    """
    Frozen class gives the ability to freeze the inheriting class attributes once an object is initialized

    Attributes
    ----------
    __frozen__: str
        attribute to freeze an object attributes after its initialization
    """
    __frozen__ = False

    def __setattr__(self, attr, value):
        if self.__frozen__ and not hasattr(self, attr):
            raise FrozenError(f"Can not set new attribute '{attr}'. "
                              f"{self.__class__.__name__} is an instance of a frozen class")
        object.__setattr__(self, attr, value)

    def _freeze(self):
        self.__frozen__ = True



import logging
from xtb_config.constants import LOG_LEVEL

logger = logging.getLogger("xtb-broker")
logger.setLevel(getattr(logging, LOG_LEVEL))

FORMAT = '[%(asctime)-15s][%(thread)d][%(levelname)s][%(module)s::%(funcName)s:%(lineno)d] - %(message)s'
logging.basicConfig(format=FORMAT)


